<?php

include_once __DIR__ .'/classes/Transaction.php';

class FirstData extends PaymentModule {

    public static $fserv_url = 'https://www.ipg-online.com/connect/gateway/processing';
    //public static $fserv_url = 'https://test2.ipg-online.com/connect/gateway/processing';
    
	protected		$_html			= '';
	protected		$_errors		= array();
	protected		$posid			= null;
	protected		$sharedKey		= null;
	protected		$testMode		= true;
	protected		$logFile		= null;
	protected		$_languages		= null;
	
	public			$token			= null;
	public			$fields_list	= null;
	public			$avaliableCurrencies = array('PLN', 'EUR');
	
	public function __construct() {
		$this->name = 'firstdata';
		$this->tab = 'payments_gateways';
		$this->version = '1.2.1';
		$this->author = 'OpenNet.pl';
		$this->controllers = array('payment', 'validation');
		$this->bootstrap = true;
        $this->is_eu_compatible = true;
        $this->ps_versions_compliancy = [
            'min' => '1.6',
            'max' => _PS_VERSION_
        ];

		parent::__construct();

		# get params
		$config = Configuration::getMultiple(array(
			'FIRSTDATA_POSID',
			'FIRSTDATA_SHAREDKEY',
			'FIRSTDATA_TESTMODE',
			'FIRSTDATA_LOGFILE',
		));
		$this->posid = isset($config['FIRSTDATA_POSID']) ? $config['FIRSTDATA_POSID'] : null;
		$this->sharedKey = isset($config['FIRSTDATA_SHAREDKEY']) ? $config['FIRSTDATA_SHAREDKEY'] : null;
		$this->testMode = isset($config['FIRSTDATA_TESTMODE']) ? $config['FIRSTDATA_TESTMODE'] : false;
		$this->logFile = _PS_ROOT_DIR_ .'/var/logs/firstdata_' . @$config['FIRSTDATA_LOGFILE'] .'.txt';

		$this->displayName = $this->l('FirstData payment gateway');
		
		if(empty($this->posid) || empty($this->sharedKey))
			$this->warning = $this->l('Account details must be configured before using this module.');
	}
	
	public function install() {
		$this->_languages = Language::getLanguages(false);
	
		# std
		if(	!parent::install() 
			|| !$this->registerHook('adminOrder')
			|| !$this->registerHook('displayPayment')
			|| !$this->registerHook('paymentOptions')
			|| !$this->registerHook('paymentReturn')
			)
			return false;
		
		# add states
		$state1 = new OrderState();
		$state1->module_name = $this->name;
		$state1->name = array();
		foreach($this->_languages as $lang)
			$state1->name[(int)$lang['id_lang']] = $this->l('Payment by FirstData - select method');
		$state1->color = '#4169E1';
		$state1->add();
		$state2 = new OrderState();
		$state2->module_name = $this->name;
		$state2->name = array();
		foreach($this->_languages as $lang)
			$state2->name[(int)$lang['id_lang']] = $this->l('Payment by FirstData - pending');
			$state2->color = '#4169E1';
			$state2->add();
		
		# copy state icon
		$source = dirname(__FILE__).'/views/img/logo.gif';
		$destination = dirname(__FILE__).'/../../img/os/'.(int)$state1->id.'.gif';
		copy($source, $destination);
		$destination = dirname(__FILE__).'/../../img/os/'.(int)$state2->id.'.gif';
		copy($source, $destination);
		
		# save state id
		Configuration::updateValue('FIRSTDATA_ORDERSTATE', $state1->id);
		Configuration::updateValue('FIRSTDATA_ORDERSTATE_PENDING', $state2->id);
		
		# create log file
		$logFile = uniqid();
		Configuration::updateValue('FIRSTDATA_LOGFILE', $logFile);
		
		# create transaction table
		Db::getInstance()->query("
			CREATE TABLE IF NOT EXISTS `"._DB_PREFIX_."firstdata_transaction` (
			  `id_transaction` int(10) unsigned NOT NULL AUTO_INCREMENT,
			  `id_order` int(10) unsigned NOT NULL,
			  `status` int(10) unsigned NOT NULL,
			  `error` int(10) unsigned NOT NULL,
			  `msg` VARCHAR( 1024 ) NULL,
			  `fd_tid` varchar(50) NOT NULL,
              `card_number` varchar(50) NOT NULL,
              `card_brand` varchar(20) NOT NULL,
              `card_expiration` varchar(10) NOT NULL,
              `card_holder` varchar(50) NOT NULL,
			  `date_add` datetime NOT NULL,
			  `date_mod` datetime DEFAULT NULL,
			  `date_accept` datetime DEFAULT NULL,
			  `date_cancel` datetime DEFAULT NULL,
			  PRIMARY KEY (`id_transaction`)
			) ENGINE=InnoDB  DEFAULT CHARSET=utf8 AUTO_INCREMENT=1;
		");
			
		return true;
	}
	
	public function uninstall() {
		$id_state1 = (int) Configuration::get('FIRSTDATA_ORDERSTATE');
		$id_state2 = (int) Configuration::get('FIRSTDATA_ORDERSTATE_PENDING');
		
		if(	!Configuration::deleteByName('FIRSTDATA_ORDERSTATE')
			|| ($id_state2 && !Configuration::deleteByName('FIRSTDATA_ORDERSTATE_PENDING'))
			|| !Configuration::deleteByName('FIRSTDATA_POSID')
			|| !Configuration::deleteByName('FIRSTDATA_SHAREDKEY')
			|| !Configuration::deleteByName('FIRSTDATA_TESTMODE')
			|| !parent::uninstall() )
			return false;
			
		$state1 = new OrderState($id_state1);
		$state1->delete();
		if($id_state2) {
			$state2 = new OrderState($id_state2);
			$state2->delete();
		}
		
		Db::getInstance()->query("DROP TABLE IF EXISTS `"._DB_PREFIX_."firstdata_transaction`");
		
		return true;
	}
	
	public function getSharedKey() {
	    return $this->sharedKey;
	}

	public function getTransactionStatuses() {
		return array(
			0 => $this->l('AUTHORIZATION_READY'),
		    1 => $this->l('CLEARING_READY'),
		    2 => $this->l('AUTHORIZED'),
			9 => $this->l('CANCELED'),
		    99 => $this->l('REJECTED')
		);
	}

	protected function _postprocess() {
		if(Tools::isSubmit('saveConfig'))
		{
			$posId = trim(Tools::getValue('FIRSTDATA_POSID'));
			Configuration::updateValue('FIRSTDATA_POSID', $posId);
			$sharedKey = trim(Tools::getValue('FIRSTDATA_SHAREDKEY'));
			Configuration::updateValue('FIRSTDATA_SHAREDKEY', $sharedKey);
			
			$testMode = trim(Tools::getValue('FIRSTDATA_TESTMODE'));
			Configuration::updateValue('FIRSTDATA_TESTMODE', $testMode);
			$this->_html .= $this->displayConfirmation($this->l('Settings updated'));
		}
		if(Tools::isSubmit('firstdata_refresh'))
		{
			$data = FirstDataTransaction::getPendingList();
			$err = false;
			if($data) {
				foreach ($data as $r) {
					$resp = $this->getTrans($r['id_order'], substr($r['date_add'], 0, 10));
					if(is_string($resp)) {
						$err = $resp;
						break;
					}

					if(($resp->orderCode == $r['id_order']) && ($resp->paymentStatus != $r['status'])) {	// Update transaction data
						$order = new Order($r['id_order']);
						$this->updateTrans($order, $resp->paymentStatus, '');						
					}
				}
			}
			
			$data = FirstDataTransaction::getNotRegisteredList(30);
			if($data) {
				foreach ($data as $r) {
					$order = new Order($r['id_order']);
					$bro = $order->getBrother();
					$hasBro = false;
					foreach($bro as $b) {
						$id_trans = FirstDataTransaction::getIdByOId($b->id);
						if($id_trans) {
							$hasBro = true;
							break;
						}
					}
					if($hasBro) continue;

					$resp = $this->getTrans($r['id_order'], substr($r['date_add'], 0, 10));
					if(!is_string($resp)) $stat = $resp->paymentStatus;
					else $stat = 999;
					$this->updateTrans($order, $stat, '');
				}
			}
				
			if(!$err) $this->_html .= $this->displayConfirmation($this->l('Transaction list updated'));
			else $this->_html .= $this->displayError($this->l($err));
		}
	}
	
	function callApi($req) {
		if(!$this->user || !$this->pass) return $this->l('User/password not set');
		if(!extension_loaded ('curl')) return $this->l('CURL extension not loaded');
		$url = self::REST_ADDR . $req;
		$curl = curl_init ();
		curl_setopt ( $curl, CURLOPT_URL, $url );
		curl_setopt ( $curl, CURLOPT_RETURNTRANSFER, 1 );
		curl_setopt ( $curl, CURLOPT_FAILONERROR, false );
		curl_setopt ( $curl, CURLOPT_USERPWD, $this->mid . '.' . $this->user . ":" . $this->pass);
	
		$result = curl_exec ( $curl );
		if($result === false) {
			$stat = curl_error($curl);
			curl_close ( $curl );
			return $stat;
		}
		$info = curl_getinfo($curl);
		if($info['http_code'] >= 400) {
			$stat = null;
			if($result) {
				$json = json_decode($result);
				$stat = $json->fault;
			}
			if($stat) $resp = $stat;
			elseif($info['http_code'] == 401) $resp = $this->l('Auth error');
			else $resp = 'HTTP error ' . $info['http_code'];
			curl_close ( $curl );
			return $resp;
		}
		curl_close ( $curl );
		return json_decode($result);
	}

	public function getContent() {
		$this->token = Tools::getAdminTokenLite('AdminModules');
		$this->_languages = Language::getLanguages(false);
		$this->_html = '';
		
		$this->_postprocess();
		if($this->_errors) {
			foreach($this->_errors as $err)
				$this->_html .= $this->displayError($err);
			$this->_html .= '<br />';
		}

		$this->_html .= $this->display(__FILE__, 'config.tpl');
		$this->_html .= $this->_displayConfigForm();
		$this->_html .= $this->_displayTransactionList();
		$this->_html .= '';

		return $this->_html;
	}

	public function _displayConfigForm() {
		$fields_form = array(
			'form' => array(
				'legend' => array(
					'title' => $this->l('Account details'),
					'icon' => 'icon-cogs'
				),
				'input' => array(
					array(
						'type' => 'text',
						'label' => $this->l('Shop Id'),
						'name' => 'FIRSTDATA_POSID',
					),
					array(
						'type' => 'text',
						'label' => $this->l('Shared Key'),
						'name' => 'FIRSTDATA_SHAREDKEY',
					),
				),
				'submit' => array(
					'title' => $this->l('Save'),
				)
			),
		);
		
		$helper = new HelperForm();
		$helper->bootstrap = true;
		$helper->className = 'bootstrap';
		$helper->show_toolbar = false;
		$helper->table = $this->table;
		
		$lang = new Language((int)Configuration::get('PS_LANG_DEFAULT'));
		$helper->default_form_language = $lang->id;
		$helper->allow_employee_form_lang = Configuration::get('PS_BO_ALLOW_EMPLOYEE_FORM_LANG') ? Configuration::get('PS_BO_ALLOW_EMPLOYEE_FORM_LANG') : 0;
		
		$this->fields_form = array();
		$helper->id = 'firstdata';
		$helper->identifier = $this->identifier;
		$helper->submit_action = 'saveConfig';
		$helper->currentIndex = $this->context->link->getAdminLink('AdminModules', false).'&configure='.$this->name.'&tab_module='.$this->tab.'&module_name='.$this->name;
		$helper->token = $this->token;
		$helper->tpl_vars = array(
			'fields_value' => array(
				'FIRSTDATA_POSID' => Tools::getValue('FIRSTDATA_POSID', Configuration::get('FIRSTDATA_POSID')),
				'FIRSTDATA_SHAREDKEY' => Tools::getValue('FIRSTDATA_SHAREDKEY', Configuration::get('FIRSTDATA_SHAREDKEY')),
				'FIRSTDATA_TESTMODE' => Tools::getValue('FIRSTDATA_TESTMODE', Configuration::get('FIRSTDATA_TESTMODE')),
			),
			'languages' => $this->context->controller->getLanguages(),
			'id_language' => $this->context->language->id
		);

		return $helper->generateForm(array($fields_form));
	}

	protected function _displayTransactionList() {
		$status = $this->getTransactionStatuses();

		$savedParams = $this->context->cookie->firstdatalist 
			? $this->context->cookie->firstdatalist 
			: array();
		
		if(Tools::isSubmit('orderBy'))
			$savedParams['orderBy'] = Tools::getValue('orderBy');
		if(Tools::isSubmit('orderWay'))
			$savedParams['orderWay'] = Tools::getValue('orderWay');
		if(Tools::isSubmit('enum'))
			$savedParams['enum'] = Tools::getValue('enum');

		$savedParams = $this->context->cookie->firstdatalist;
		
		$orderBy = $savedParams['orderBy'] ? $savedParams['orderBy'] : 'date_add';
		$orderWay = $savedParams['orderWay'] ? $savedParams['orderBy'] : 'DESC';
		$page = Tools::getValue('page', 1);
		$enum = (int) ($savedParams['enum'] ? $savedParams['enum'] : 25);
		
		$data = FirstDataTransaction::getList(
			array(),
			$orderBy, $orderWay,
			$page, $enum
			);
	
		if(!$data['total'])
			return null;
	
		$pagesNum = ceil($data['total'] / $enum);
	
		# assign
		$this->context->smarty->assign(array(
			'baseLink' => $this->context->link->getAdminLink('AdminModules', false).'&configure='.$this->name.'&tab_module='.$this->tab.'&module_name='.$this->name,
		
			'orderBy' => $orderBy,
			'orderWay' => $orderWay,
			'page' => $page,
			'enum' => $enum,
			'pagesNum' => $pagesNum,
			
			'status' => $status,
			
			'data' => $data['records'],
		));
		
		return $this->display(__FILE__, 'trans_list.tpl');
	}

    public function hookPaymentOptions($params) {
        if (!$this->active)
            return;

        if (!$this->checkCurrency())
            return;

        $newOption = new PrestaShop\PrestaShop\Core\Payment\PaymentOption();
        $newOption->setModuleName($this->name)
                ->setCallToActionText($this->l('Pay online by FirstData'))
                ->setAction($this->context->link->getModuleLink($this->name, 'payment', array(), true))
                ->setLogo( Tools::getShopDomainSsl(true, true).__PS_BASE_URI__.'modules/'.$this->name.'/views/img/paymentlogo_min.png' )
                ->setAdditionalInformation($this->display(__FILE__, 'payment_option.tpl'));

        return array(
            $newOption,
		);
    }

	public function hookDisplayPayment($params) {
		if(!$this->active)
			return;

		$this->smarty->assign(array(
			'this_path' => $this->_path,
			'this_path_bw' => $this->_path,
			'this_path_ssl' => Tools::getShopDomainSsl(true, true) . __PS_BASE_URI__ . 'modules/' . $this->name . '/'
		));
		
		$ps16 = strpos(_PS_VERSION_, '1.6') === 0;
		if($ps16) $tpl = 'payment_16.tpl';
        else $tpl = 'payment_17.tpl';
		return $this->display(__FILE__, $tpl);
	}

	public function hookAdminOrder($params) {
		$order = new Order((int)$params['id_order']);
		$status = '--';
		$linked = '';
		if($order->module  == 'firstdata') {
			$id_trans = FirstDataTransaction::getIdByOId($order->id);
//			error_log('hookAdminOrder='.$id_trans);
			if($id_trans == null) {
				$bro = $order->getBrother();
				foreach($bro as $b) {
					$id_trans = FirstDataTransaction::getIdByOId($b->id);
					if($id_trans) {
						$linked = ' (główne zamówienie powiązane: ' . $b->id . ')';
						break;
					}
				}
			}
			if($id_trans) {
				$trans = new FirstDataTransaction($id_trans);
				$slist = $this->getTransactionStatuses();
				if(isset($slist[$trans->status])) $status = $slist[$trans->status];
				else $status = '???';
			}
		}
		$this->smarty->assign(array(
			'is_firstdata' => ($order->module == 'firstdata'),
			'order_id' => $order->id,
			'key' => md5($order->id . Configuration::get('FIRSTDATA_SHAREDKEY')),
			'status' => $status,
			'url' => _MODULE_DIR_ . 'firstdata/ajax.php',
			'firstdata_token' => sha1(_COOKIE_KEY_.$this->name),
			'linked' => $linked
		));
		return $this->display(__FILE__, 'adminOrder.tpl');
	}

	public function doAction($action, $id, $key) {
		if(md5($id . Configuration::get('FIRSTDATA_SHAREDKEY')) != $key) return array('error' => $this->l('Security error'));
		switch ($action) {
			case 'R':
				return $this->reload($id);
				break;
			default:
				return array('error' => $this->l('Internal error'));
		}
	}
	
	private function reload($id) {
		$order = new Order($id);
		$oid = $id;
		$id_trans = FirstDataTransaction::getIdByOId($order->id);
		if($id_trans == null) {
			$bro = $order->getBrother();
			foreach($bro as $b) {
				$id_trans = FirstDataTransaction::getIdByOId($b->id);
				if($id_trans) {
					$oid = $b->id;
					break;
				}
			}
		}
		$status_o = 0;
		if($id_trans) {
			$trans = new FirstDataTransaction($id_trans);
			$status_o = $trans->status;
		}

		$resp = $this->getTrans($oid, substr($order->date_add, 0, 10));
		if(is_string($resp)) return array('error' => $resp);
		if($resp->orderCode != $oid) return array('error' => 'id error');
		$status_n = $resp->paymentStatus;
		 
		if($status_n != $status_o) {
			$this->updateTrans($order, $status_n, '');
			return array('reload' => 1, 'info' => $this->l('Status has changed, the page will reload'));
		}
		return array('info' => $this->l('Status not changed'));
	}

	private function getTrans($id, $d) {
		$api_resp = $this->callApi('/merchant/' . $this->mid . '/payment/' . $id . ';date=' .$d . '?format=json');
		
		if(!is_array($api_resp)) return $api_resp;
		$resp = $api_resp[count($api_resp) - 1];

		return $resp;		
	}

    public function hookPaymentReturn($params) {
        if(!$this->active)
            return;

		$result = Tools::getValue('result');
		$order = null;
		if(isset($params['order'])) {
			$order = $params['order'];
		}
		else {
			$orderId = (int) Tools::getValue('id_order');
			$order = new Order($orderId);
		}

        $status = null;
        
        $id_trans = FirstDataTransaction::getIdByOId($order->id);
        if($id_trans != null) {
            $trans = new FirstDataTransaction($id_trans);
            $msg = json_decode($trans->msg, true);
            $msg['strcurrency'] = ($msg['currency'] == 985 ? 'PLN' : ($msg['currency'] == 978 ? 'EUR' : '???'));
        }
        else $msg = array();

		if($result == 'error')
			$status = 'error';
		elseif($order->current_state == Configuration::get('PS_OS_PAYMENT'))
			$status = 'paid';
		elseif($order->current_state == Configuration::get('FIRSTDATA_ORDERSTATE'))
			$status = 'fd_new';
		elseif($order->current_state == Configuration::get('FIRSTDATA_ORDERSTATE_PENDING'))
			$status = 'fd_pending';
		elseif($order->current_state == Configuration::get('PS_OS_ERROR'))
			$status = 'error';
		elseif($order->current_state == Configuration::get('PS_OS_CANCELED'))
			$status = 'canceled';
        else
			$status = 'pending';
//			error_log('hookPaymentReturn r=' . $result . ' s=' . $status . ' c=' . $order->current_state . ' o=' . $order->id . ' t=' . $id_trans);
//		error_log('return msg='.print_r($msg, true));

		if(md5($order->id . $this->getSharedKey()) == Tools::getValue('control')) {
		    $url = Context::getContext()->link->getModuleLink('firstdata', 'payment');
			$data = array(
				'url'		=> $url,
			    'order_id'  => $order->id,
			    'post' => $msg
			);
		
		}
		else $data = array('order_id' => 0);

		$data['status'] = $status;
		$data['contact_url'] = $this->context->link->getPageLink('contact', true);		
		$this->smarty->assign($data);

		return $this->display(__FILE__, 'payment_return.tpl');
    }
	
	public function checkCurrency() {
		$currency = new Currency($this->context->cart->id_currency);
		return in_array($currency->iso_code, $this->avaliableCurrencies);
	}
	
	public function validation() {
//	    error_log('valid post='.print_r($_POST,true));	
		$cart = $this->context->cart;
		if(!Validate::isLoadedObject($cart) || Order::getOrderByCartId($cart->id))
			return false;

		$currency = new Currency($cart->id_currency);
		$customer = $this->context->customer;
		$orderTotal = (float) $cart->getOrderTotal(true, Cart::BOTH);
		
		$result = $this->validateOrder(
			$cart->id,
			Configuration::get('FIRSTDATA_ORDERSTATE'),
			$orderTotal,
			$this->displayName,
			null,
			array(),
			(int) $currency->id,
			false,
			$customer->secure_key
		);
		if(!$result)
			return false;

		$order = new Order($this->currentOrder);
		$address = new Address($order->id_address_delivery);
		$country = new Country($address->id_country);
		$myurl = Tools::getShopDomainSsl(true, true).__PS_BASE_URI__.'modules/'.$this->name.'/';
		$lang = $this->context->language->iso_code == 'pl' ? 'pl_PL' : 'en_US';
		
		$data = array(
		    'chargetotal' => round($orderTotal, 2),
		    'currency' => $currency->iso_code_num,
		    'checkoutoption' => 'combinedpage',
		    'hash_algorithm' => 'HMACSHA256',
		    'language' => $lang,
		    'oid' => $order->reference,
		    'responseFailURL' => $myurl . 'error.php',
		    'responseSuccessURL' => $myurl . 'success.php',
		    'storename' => $this->posid,
		    'timezone' => Configuration::get('PS_TIMEZONE'),
		    'transactionNotificationURL' => $myurl . 'request.php',
		    'txndatetime' => date("Y:m:d-H:i:s"),
		    'txntype' => 'sale',
			'bcity' => trim($address->city),
			'bzip' => trim($address->postcode),
			'bcountry' => trim($country->iso_code),
			'email' => trim($customer->email)
		);
		
		ksort($data);
		$data['hashExtended'] = $this->createExtendedHash($data);
//		error_log('valid data='.print_r($data,true));	
		return $data;
	}
	
    public function request($notification = false) {
		if(!$_POST) {
			$this->_log('no input');
			return false;
		}
//        error_log('request post=' . print_r($_POST,true));
		$error = false;
		$acode = explode(':', $_POST['approval_code']);
		$status = $acode[0];

		if($notification) {
		    $data = array(
		        'chargetotal' => $_POST['chargetotal'],
		        'currency' => $_POST['currency'],
		        'txndatetime' => $_POST['txndatetime'],
		        'storename' => trim($this->posid),
		        'approval_code' => $_POST['approval_code']
		    );
		    $hash = $_POST['notification_hash'];
		}
		else {
		    $data = array(
		        'approval_code' => $_POST['approval_code'],
		        'chargetotal' => $_POST['chargetotal'],
		        'currency' => $_POST['currency'],
		        'txndatetime' => $_POST['txndatetime'],
		        'storename' => trim($this->posid)
		    );
		    $hash = $_POST['response_hash'];
		}
		$chash = $this->createExtendedHash($data);
		if($hash != $chash) {
		    $text = "Wrong controlData in confirmation request.\r\n  order ref: ".$_POST['oid'];
			if(isset($data['ipgTransactionId']))
			    $text .= "\r\n  Transadtion id: ".$_POST['ipgTransactionId'];
			$this->_log($text);
			
			return false;
		}

		$data['chargetotal'] = 1 * str_replace(',', '.', $data['chargetotal']);
			
		$order_col = Order::getByReference($_POST['oid']);
		$order = $order_col->getFirst();
		$bro = $order->getBrother();
		
		$total = $order->total_paid;
		foreach($bro as $b) $total += $b->total_paid;
		if($data['chargetotal'] != $total) {
			$state = (int) Configuration::get('PS_OS_ERROR');
			$this->updOrderStatus($order, $bro, $state);

			$text = "Wrong payment amount.\r\n  order ref: ".$_POST['oid'];
			if(isset($data['ipgTransactionId']))
				$text .= "\r\n  Transadtion id: ".$data['ipgTransactionId'];
			$this->_log($text);

			return false;
		}

		if($status == 'N') {
			$text = "Payment error.\r\n  error code: " . $error . "\r\n  msg: " . $acode[2] . "\r\n  order ref: " . $_POST['oid'];
			if(isset($_POST['ipgTransactionId']))
			    $text .= "\r\n  Transadtion id: ".$_POST['ipgTransactionId'];
			$this->_log($text);
			$this->updateTrans($order, $status, $data['chargetotal']);
			
			return false;
		}
		
		if(!$this->updateTrans($order, $status, $data['chargetotal'])) return false;

		$text = "Transaction complete\r\n  Transadtion id: ". $_POST['ipgTransactionId'];
		$this->_log($text);
		
		return true;
	}

	private function updOrderStatus($order, $bro, $state) {
		if($order->current_state != $state) {
			$order->setCurrentState($state);
			foreach($bro as $b) {
			    $b->setCurrentState($state);
			}
		}
	}

	private function updateTrans($order, $status, $amount) {
		$bro = $order->getBrother();
		$transaction_id = isset($_POST['ipgTransactionId']) ? $_POST['ipgTransactionId'] : '';
		$id_trans = FirstDataTransaction::getIdByOId($order->id);
//		$id_trans = FirstDataTransaction::getIdByFDId($transaction_id);

		if($id_trans) {
			$trans = new FirstDataTransaction($id_trans);
		}
		else {
			$trans = new FirstDataTransaction();
			$trans->id_order = $order->id;
			$trans->status = 0;
			$trans->error = 0;
			$trans->fd_tid = $transaction_id;
			$trans->card_number = isset($_POST['cardnumber']) ? $_POST['cardnumber'] : '';
			$trans->card_brand = $_POST['ccbrand'];
			$trans->card_expiration = isset($_POST['expmonth']) && isset($_POST['expyear']) ? ($_POST['expmonth'] . '/' . $_POST['expyear']) : '';
			$trans->card_holder = isset($_POST['bname']) ? $_POST['bname'] : '';
			if(!$trans->add()) {
				$state = (int) Configuration::get('PS_OS_ERROR');
				$this->updOrderStatus($order, $bro, $state);
				$text = "Can not create transaction in DB.\r\n  order id: " . $order->id;
				if(isset($transaction_id))
					$text .= "\r\n  Transadtion id: ".$transaction_id;
				$this->_log($text);
							
				return false;
			}
		}

		$msg = array(
		    'status' => $status,
		    'txndate_processed' => isset($_POST['txndate_processed']) ? $_POST['txndate_processed'] : '',
		    'approval_code' => $_POST['approval_code'],
		    'chargetotal' => $_POST['chargetotal'],
		    'currency' => $_POST['currency'],
		);
		
		if($status == 'Y') {
			$trans->status = 2;
			$trans->date_accept = date('Y-m-d H:i:s');
			$trans->error = 0;
			$msg['ipgTransactionId'] = $_POST['ipgTransactionId'];
			$trans->msg = json_encode($msg);
			
			$state = (int) Configuration::get('PS_OS_PAYMENT');
			$this->updOrderStatus($order, $bro, $state);
		}
		elseif($status == '?') {
			$trans->status = 1;
			$msg['ipgTransactionId'] = $_POST['ipgTransactionId'];
			$trans->msg = json_encode($msg);
			
			$state = $this->getPendingOs();
			$this->updOrderStatus($order, $bro, $state);
		}
		else {
		    $trans->status = 9;
		    $trans->date_cancel  = date('Y-m-d H:i:s');
		    $error = isset($_POST['fail_rc']) ? $_POST['fail_rc'] : 0;
		    $trans->error = $error;
		    $msg['fail_reason'] = $_POST['fail_reason'];
		    $trans->msg = json_encode($msg);

		    $state = ($error == 5993)
		    ? (int) Configuration::get('PS_OS_CANCELED')
		    : (int) Configuration::get('PS_OS_ERROR');
		    $this->updOrderStatus($order, $bro, $state);
		}
		$result = $trans->save();
		if(!$result) {
			$state = (int) Configuration::get('PS_OS_ERROR');
			$this->updOrderStatus($order, $bro, $state);
			$text = "Can not update transaction in DB.\r\n  order id: " . $order->id;
			if(isset($transaction_id))
				$text .= "\r\n  Transadtion id: ".$transaction_id;
			$this->_log($text);
						
			return false;
		}
		return true;
	}

	private function getPendingOs() {
		$id = Configuration::get('FIRSTDATA_ORDERSTATE_PENDING');
		if(!$id) {
			$state2 = new OrderState();
			$state2->module_name = $this->name;
			$state2->name = array();
			$this->_languages = Language::getLanguages(false);
			foreach($this->_languages as $lang)
				$state2->name[(int)$lang['id_lang']] = $this->l('Payment by FirstData - pending');
			$state2->color = '#4169E1';
			$state2->add();
			$source = dirname(__FILE__).'/views/img/logo.gif';
			$destination = dirname(__FILE__).'/../../img/os/'.(int)$state2->id.'.gif';
			copy($source, $destination);
			Configuration::updateValue('FIRSTDATA_ORDERSTATE_PENDING', $state2->id);					
		}
		return (int) $id;
	}
		
	public function createExtendedHash($data) {
	    $stringToHash = join('|', $data);
//	    error_log('s=' . $stringToHash . ' k=' . $this->sharedKey);
	    $hash = hash_hmac('sha256', $stringToHash, $this->sharedKey, true);
	    
	    return base64_encode($hash);
	}
	
	protected function _log($text) {
		$text = date('Y-m-d H:i:s') .' '. $text ."\r\n\r\n";
		
		$fp = fopen($this->logFile, 'a');
		fwrite($fp, $text);
		fclose($fp);
	}

}
